export(fd_block)
export(fd_compute)
export(fd_connect)
export(fd_define)
export(fd_fetch)
export(fd_run)
export(fd_upsert)
S3method(print,featdelta_block)
S3method(print,featdelta_con)
S3method(print,featdelta_defs)
S3method(print,fd_run_report)
S3method(summary,featdelta_defs)
