% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fd_define.R
\name{print.featdelta_defs}
\alias{print.featdelta_defs}
\title{Print a featdelta definitions object}
\usage{
\method{print}{featdelta_defs}(x, ...)
}
\arguments{
\item{x}{A \code{featdelta_defs} object.}

\item{...}{Unused.}
}
\value{
The input object invisibly.
}
\description{
Prints a concise human-readable summary of a \code{featdelta_defs} object.
}
\seealso{
Other featdelta defs helpers: 
\code{\link{fd_block}()},
\code{\link{fd_define}()},
\code{\link{summary.featdelta_defs}()}
}
\concept{featdelta defs helpers}
