% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fd_define.R
\name{summary.featdelta_defs}
\alias{summary.featdelta_defs}
\title{Summarize a featdelta definitions object}
\usage{
\method{summary}{featdelta_defs}(object, ...)
}
\arguments{
\item{object}{A \code{featdelta_defs} object.}

\item{...}{Unused.}
}
\value{
A data frame with one row per definition step and columns such as
step name, type, mode, declared outputs, and stored expression text.
}
\description{
Returns a compact data-frame summary of the ordered definition steps stored
in a \code{featdelta_defs} object.
}
\seealso{
Other featdelta defs helpers: 
\code{\link{fd_block}()},
\code{\link{fd_define}()},
\code{\link{print.featdelta_defs}()}
}
\concept{featdelta defs helpers}
